@@ -86,18 +86,28 @@ class DictCursorBase(_cursor):
             res = _cursor.fetchall(self)
         return res
 
-    def next(self):
+    def __iter__(self):
         if self._prefetch:
-            res = _cursor.fetchone(self)
-            if res is None:
-                raise StopIteration()
+            res = _cursor.fetchmany(self, self.itersize)
+            if not res:
+                return
         if self._query_executed:
             self._build_index()
         if not self._prefetch:
-            res = _cursor.fetchone(self)
-            if res is None:
-                raise StopIteration()
-        return res
+            res = _cursor.fetchmany(self, self.itersize)
+
+        for r in res:
+            yield r
+
+        # the above was the first itersize record. the following are
+        # in a repeated loop.
+        while 1:
+            res = _cursor.fetchmany(self, self.itersize)
+            if not res:
+                return
+            for r in res:
+                yield r
+
 
 class DictConnection(_connection):
     """A connection that uses `DictCursor` automatically."""
@@ -694,7 +704,7 @@ WHERE typname = 'hstore';
 
         # revert the status of the connection as before the command
         if (conn_status != _ext.STATUS_IN_TRANSACTION
-        and conn.isolation_level != _ext.ISOLATION_LEVEL_AUTOCOMMIT):
+        and not conn.autocommit):
             conn.rollback()
 
         return tuple(rv0), tuple(rv1)
@@ -831,8 +841,8 @@ class CompositeCaster(object):
         tokens = self.tokenize(s)
         if len(tokens) != len(self.atttypes):
             raise psycopg2.DataError(
-                "expecting %d components for the type %s, %d found instead",
-                (len(self.atttypes), self.name, len(self.tokens)))
+                "expecting %d components for the type %s, %d found instead" %
+                (len(self.atttypes), self.name, len(tokens)))
 
         attrs = [ curs.cast(oid, token)
             for oid, token in zip(self.atttypes, tokens) ]
@@ -903,7 +913,8 @@ SELECT t.oid, %s, attname, atttypid
 FROM pg_type t
 JOIN pg_namespace ns ON typnamespace = ns.oid
 JOIN pg_attribute a ON attrelid = typrelid
-WHERE typname = %%s and nspname = %%s
+WHERE typname = %%s AND nspname = %%s
+    AND attnum > 0 AND NOT attisdropped
 ORDER BY attnum;
 """ % typarray, (tname, schema))
 
@@ -911,7 +922,7 @@ ORDER BY attnum;
 
         # revert the status of the connection as before the command
         if (conn_status != _ext.STATUS_IN_TRANSACTION
-        and conn.isolation_level != _ext.ISOLATION_LEVEL_AUTOCOMMIT):
+        and not conn.autocommit):
             conn.rollback()
 
         if not recs:
