@@ -26,11 +26,6 @@
 #ifndef PSYCOPG_CURSOR_H
 #define PSYCOPG_CURSOR_H 1
 
-#define PY_SSIZE_T_CLEAN
-#include <Python.h>
-#include <libpq-fe.h>
-
-#include "psycopg/config.h"
 #include "psycopg/connection.h"
 
 #ifdef __cplusplus
@@ -39,18 +34,20 @@ extern "C" {
 
 extern HIDDEN PyTypeObject cursorType;
 
-typedef struct {
+/* the typedef is forward-declared in psycopg.h */
+struct cursorObject {
     PyObject_HEAD
 
     connectionObject *conn; /* connection owning the cursor */
 
     int closed:1;            /* 1 if the cursor is closed */
     int notuples:1;          /* 1 if the command was not a SELECT query */
-    int needsfetch:1;        /* 1 if a call to pq_fetch is pending */
+    int withhold:1;          /* 1 if the cursor is named and uses WITH HOLD */
 
     long int rowcount;       /* number of rows affected by last execute */
     long int columns;        /* number of columns fetched from the db */
     long int arraysize;      /* how many rows should fetchmany() return */
+    long int itersize;       /* how many rows should iter(cur) fetch in named cursors */
     long int row;            /* the row counter for fetch*() operations */
     long int mark;           /* transaction marker, copied from conn */
 
@@ -62,13 +59,13 @@ typedef struct {
     PyObject   *pgstatus;  /* last message from the server after an execute */
     Oid         lastoid;   /* last oid from an insert or InvalidOid */
 
-    PyObject *casts;      /* an array (tuple) of typecast functions */
-    PyObject *caster;     /* the current typecaster object */
+    PyObject *casts;       /* an array (tuple) of typecast functions */
+    PyObject *caster;      /* the current typecaster object */
 
-    PyObject *copyfile;   /* file-like used during COPY TO/FROM ops */
+    PyObject  *copyfile;   /* file-like used during COPY TO/FROM ops */
     Py_ssize_t copysize;   /* size of the copy buffer during COPY TO/FROM ops */
 #define DEFAULT_COPYSIZE 16384
-#define DEFAULT_COPYBUFF  8132
+#define DEFAULT_COPYBUFF  8192
 
     PyObject *tuple_factory;    /* factory for result tuples */
     PyObject *tzinfo_factory;   /* factory for tzinfo objects */
@@ -82,9 +79,13 @@ typedef struct {
     PyObject *string_types;   /* a set of typecasters for string types */
     PyObject *binary_types;   /* a set of typecasters for binary types */
 
-} cursorObject;
+    PyObject *weakreflist;    /* list of weak references */
+
+};
+
 
 /* C-callable functions in cursor_int.c and cursor_ext.c */
+HIDDEN PyObject *curs_get_cast(cursorObject *self, PyObject *oid);
 HIDDEN void curs_reset(cursorObject *self);
 
 /* exception-raising macros */
@@ -99,10 +100,21 @@ if ((self)->notuples && (self)->name == NULL) {               \
     return NULL; }
 
 #define EXC_IF_NO_MARK(self) \
-if ((self)->mark != (self)->conn->mark) {                                  \
+if ((self)->mark != (self)->conn->mark && (self)->withhold == 0) {                                  \
     PyErr_SetString(ProgrammingError, "named cursor isn't valid anymore"); \
     return NULL; }
 
+#define EXC_IF_CURS_ASYNC(self, cmd) if ((self)->conn->async == 1) { \
+    PyErr_SetString(ProgrammingError, #cmd " cannot be used "        \
+    "in asynchronous mode");                                         \
+    return NULL; }
+
+#define EXC_IF_ASYNC_IN_PROGRESS(self, cmd) \
+if ((self)->conn->async_cursor != NULL) {   \
+    PyErr_SetString(ProgrammingError, #cmd " cannot be used "        \
+    "while an asynchronous query is underway");                      \
+    return NULL; }
+
 #ifdef __cplusplus
 }
 #endif
