@@ -23,13 +23,9 @@
  * License for more details.
  */
 
-#define PY_SSIZE_T_CLEAN
-#include <Python.h>
-
 #define PSYCOPG_MODULE
-#include "psycopg/config.h"
-#include "psycopg/python.h"
 #include "psycopg/psycopg.h"
+
 #include "psycopg/connection.h"
 #include "psycopg/cursor.h"
 #include "psycopg/green.h"
@@ -43,6 +39,7 @@
 #include "psycopg/adapter_qstring.h"
 #include "psycopg/adapter_binary.h"
 #include "psycopg/adapter_pboolean.h"
+#include "psycopg/adapter_pint.h"
 #include "psycopg/adapter_pfloat.h"
 #include "psycopg/adapter_pdecimal.h"
 #include "psycopg/adapter_asis.h"
@@ -70,158 +67,51 @@ HIDDEN PyObject *psycoEncodings = NULL;
 HIDDEN int psycopg_debug_enabled = 0;
 #endif
 
+/* Python representation of SQL NULL */
+HIDDEN PyObject *psyco_null = NULL;
+
+/* The type of the cursor.description items */
+HIDDEN PyObject *psyco_DescriptionType = NULL;
+
 /** connect module-level function **/
 #define psyco_connect_doc \
-"connect(dsn, ...) -- Create a new database connection.\n\n"               \
-"This function supports two different but equivalent sets of arguments.\n" \
-"A single data source name or ``dsn`` string can be used to specify the\n" \
-"connection parameters, as follows::\n\n"                                  \
-"    psycopg2.connect(\"dbname=xxx user=xxx ...\")\n\n"   \
-"If ``dsn`` is not provided it is possible to pass the parameters as\n"    \
-"keyword arguments; e.g.::\n\n"                                            \
-"    psycopg2.connect(database='xxx', user='xxx', ...)\n\n"                \
-"The full list of available parameters is:\n\n"                            \
-"- ``dbname`` -- database name (only in 'dsn')\n"                          \
-"- ``database`` -- database name (only as keyword argument)\n"             \
-"- ``host`` -- host address (defaults to UNIX socket if not provided)\n"   \
-"- ``port`` -- port number (defaults to 5432 if not provided)\n"           \
-"- ``user`` -- user name used to authenticate\n"                           \
-"- ``password`` -- password used to authenticate\n"                        \
-"- ``sslmode`` -- SSL mode (see PostgreSQL documentation)\n\n"             \
-"- ``async`` -- if the connection should provide asynchronous API\n\n"     \
-"If the ``connection_factory`` keyword argument is not provided this\n"    \
-"function always return an instance of the `connection` class.\n"          \
-"Else the given sub-class of `extensions.connection` will be used to\n"    \
-"instantiate the connection object.\n\n"                                   \
-":return: New database connection\n"                                         \
-":rtype: `extensions.connection`"
-
-static size_t
-_psyco_connect_fill_dsn(char *dsn, const char *kw, const char *v, size_t i)
-{
-    strcpy(&dsn[i], kw); i += strlen(kw);
-    strcpy(&dsn[i], v); i += strlen(v);
-    return i;
-}
+"_connect(dsn, [connection_factory], [async]) -- New database connection.\n\n"
 
 static PyObject *
 psyco_connect(PyObject *self, PyObject *args, PyObject *keywds)
 {
-    PyObject *conn = NULL, *factory = NULL;
-    PyObject *pyport = NULL;
-
-    size_t idsn=-1;
-    int iport=-1;
-    const char *dsn_static = NULL;
-    char *dsn_dynamic=NULL;
-    const char *database=NULL, *user=NULL, *password=NULL;
-    const char *host=NULL, *sslmode=NULL;
-    char port[16];
+    PyObject *conn = NULL;
+    PyObject *factory = NULL;
+    const char *dsn = NULL;
     int async = 0;
 
-    static char *kwlist[] = {"dsn", "database", "host", "port",
-                             "user", "password", "sslmode",
-                             "connection_factory", "async", NULL};
+    static char *kwlist[] = {"dsn", "connection_factory", "async", NULL};
 
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "|sssOsssOi", kwlist,
-                                     &dsn_static, &database, &host, &pyport,
-                                     &user, &password, &sslmode,
-                                     &factory, &async)) {
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "s|Oi", kwlist,
+                                     &dsn, &factory, &async)) {
         return NULL;
     }
 
-    if (pyport && PyString_Check(pyport)) {
-      PyObject *pyint = PyInt_FromString(PyString_AsString(pyport), NULL, 10);
-      if (!pyint) goto fail;
-      /* Must use PyInt_AsLong rather than PyInt_AS_LONG, because
-       * PyInt_FromString can return a PyLongObject: */
-      iport = PyInt_AsLong(pyint);
-      Py_DECREF(pyint);
-    }
-    else if (pyport && PyInt_Check(pyport)) {
-      iport = PyInt_AsLong(pyport);
-    }
-    else if (pyport != NULL) {
-      PyErr_SetString(PyExc_TypeError, "port must be a string or int");
-      goto fail;
+    Dprintf("psyco_connect: dsn = '%s', async = %d", dsn, async);
+
+    /* allocate connection, fill with errors and return it */
+    if (factory == NULL || factory == Py_None) {
+        factory = (PyObject *)&connectionType;
     }
 
-    if (iport > 0)
-      PyOS_snprintf(port, 16, "%d", iport);
-
-    if (dsn_static == NULL) {
-        size_t l = 46; /* len(" dbname= user= password= host= port= sslmode=\0") */
-
-        if (database) l += strlen(database);
-        if (host) l += strlen(host);
-        if (iport > 0) l += strlen(port);
-        if (user) l += strlen(user);
-        if (password) l += strlen(password);
-        if (sslmode) l += strlen(sslmode);
-
-        dsn_dynamic = malloc(l*sizeof(char));
-        if (dsn_dynamic == NULL) {
-            PyErr_SetString(InterfaceError, "dynamic dsn allocation failed");
-            goto fail;
-        }
-
-        idsn = 0;
-        if (database)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " dbname=", database, idsn);
-        if (host)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " host=", host, idsn);
-        if (iport > 0)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " port=", port, idsn);
-        if (user)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " user=", user, idsn);
-        if (password)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " password=", password, idsn);
-        if (sslmode)
-            idsn = _psyco_connect_fill_dsn(dsn_dynamic, " sslmode=", sslmode, idsn);
-
-        if (idsn > 0) {
-            dsn_dynamic[idsn] = '\0';
-            memmove(dsn_dynamic, &dsn_dynamic[1], idsn);
-        }
-        else {
-            PyErr_SetString(InterfaceError, "missing dsn and no parameters");
-            goto fail;
-        }
-    }
-
-    {
-      const char *dsn = (dsn_static != NULL ? dsn_static : dsn_dynamic);
-      Dprintf("psyco_connect: dsn = '%s', async = %d", dsn, async);
-
-      /* allocate connection, fill with errors and return it */
-      if (factory == NULL) factory = (PyObject *)&connectionType;
-      /* Here we are breaking the connection.__init__ interface defined
-       * by psycopg2. So, if not requiring an async conn, avoid passing
-       * the async parameter. */
-      /* TODO: would it be possible to avoid an additional parameter
-       * to the conn constructor? A subclass? (but it would require mixins
-       * to further subclass) Another dsn parameter (but is not really
-       * a connection parameter that can be configured) */
-      if (!async) {
+    /* Here we are breaking the connection.__init__ interface defined
+     * by psycopg2. So, if not requiring an async conn, avoid passing
+     * the async parameter. */
+    /* TODO: would it be possible to avoid an additional parameter
+     * to the conn constructor? A subclass? (but it would require mixins
+     * to further subclass) Another dsn parameter (but is not really
+     * a connection parameter that can be configured) */
+    if (!async) {
         conn = PyObject_CallFunction(factory, "s", dsn);
-      } else {
+    } else {
         conn = PyObject_CallFunction(factory, "si", dsn, async);
-      }
     }
 
-    goto cleanup;
-    fail:
-      assert (PyErr_Occurred());
-      if (conn != NULL) {
-        Py_DECREF(conn);
-        conn = NULL;
-      }
-      /* Fall through to cleanup: */
-    cleanup:
-      if (dsn_dynamic != NULL) {
-        free(dsn_dynamic);
-      }
-
     return conn;
 }
 
@@ -233,7 +123,7 @@ psyco_connect(PyObject *self, PyObject *args, PyObject *keywds)
 "  * `conn_or_curs`: A connection, cursor or None"
 
 #define typecast_from_python_doc \
-"new_type(oids, name, adapter) -> new type object\n\n" \
+"new_type(oids, name, castobj) -> new type object\n\n" \
 "Create a new binding object. The object can be used with the\n" \
 "`register_type()` function to bind PostgreSQL objects to python objects.\n\n" \
 ":Parameters:\n" \
@@ -241,9 +131,18 @@ psyco_connect(PyObject *self, PyObject *args, PyObject *keywds)
 "  * `name`: Name for the new type\n" \
 "  * `adapter`: Callable to perform type conversion.\n" \
 "    It must have the signature ``fun(value, cur)`` where ``value`` is\n" \
-"    the string representation returned by PostgreSQL (`None` if ``NULL``)\n" \
+"    the string representation returned by PostgreSQL (`!None` if ``NULL``)\n" \
 "    and ``cur`` is the cursor from which data are read."
 
+#define typecast_array_from_python_doc \
+"new_array_type(oids, name, baseobj) -> new type object\n\n" \
+"Create a new binding object to parse an array.\n\n" \
+"The object can be used with `register_type()`.\n\n" \
+":Parameters:\n" \
+"  * `oids`: Tuple of ``oid`` of the PostgreSQL types to convert.\n" \
+"  * `name`: Name for the new type\n" \
+"  * `baseobj`: Adapter to perform type conversion of a single array item."
+
 static void
 _psyco_register_type_set(PyObject **dict, PyObject *type)
 {
@@ -289,20 +188,35 @@ static void
 psyco_adapters_init(PyObject *mod)
 {
     PyObject *call;
-    PyTypeObject *type;
 
     microprotocols_add(&PyFloat_Type, NULL, (PyObject*)&pfloatType);
-    microprotocols_add(&PyInt_Type, NULL, (PyObject*)&asisType);
-    microprotocols_add(&PyLong_Type, NULL, (PyObject*)&asisType);
+#if PY_MAJOR_VERSION < 3
+    microprotocols_add(&PyInt_Type, NULL, (PyObject*)&pintType);
+#endif
+    microprotocols_add(&PyLong_Type, NULL, (PyObject*)&pintType);
     microprotocols_add(&PyBool_Type, NULL, (PyObject*)&pbooleanType);
 
+    /* strings */
+#if PY_MAJOR_VERSION < 3
     microprotocols_add(&PyString_Type, NULL, (PyObject*)&qstringType);
+#endif
     microprotocols_add(&PyUnicode_Type, NULL, (PyObject*)&qstringType);
-    microprotocols_add(&PyBuffer_Type, NULL, (PyObject*)&binaryType);
-    microprotocols_add(&PyList_Type, NULL, (PyObject*)&listType);
 
-    if ((type = (PyTypeObject*)psyco_GetDecimalType()) != NULL)
-        microprotocols_add(type, NULL, (PyObject*)&pdecimalType);
+    /* binary */
+#if PY_MAJOR_VERSION < 3
+    microprotocols_add(&PyBuffer_Type, NULL, (PyObject*)&binaryType);
+#else
+    microprotocols_add(&PyBytes_Type, NULL, (PyObject*)&binaryType);
+#endif
+
+#if PY_MAJOR_VERSION >= 3 || PY_MINOR_VERSION >= 6
+    microprotocols_add(&PyByteArray_Type, NULL, (PyObject*)&binaryType);
+#endif
+#if PY_MAJOR_VERSION >= 3 || PY_MINOR_VERSION >= 7
+    microprotocols_add(&PyMemoryView_Type, NULL, (PyObject*)&binaryType);
+#endif
+
+    microprotocols_add(&PyList_Type, NULL, (PyObject*)&listType);
 
     /* the module has already been initialized, so we can obtain the callable
        objects directly from its dictionary :) */
@@ -317,10 +231,16 @@ psyco_adapters_init(PyObject *mod)
 
 #ifdef HAVE_MXDATETIME
     /* as above, we use the callable objects from the psycopg module */
-    call = PyMapping_GetItemString(mod, "TimestampFromMx");
-    microprotocols_add(mxDateTime.DateTime_Type, NULL, call);
-    call = PyMapping_GetItemString(mod, "TimeFromMx");
-    microprotocols_add(mxDateTime.DateTimeDelta_Type, NULL, call);
+    if (NULL != (call = PyMapping_GetItemString(mod, "TimestampFromMx"))) {
+        microprotocols_add(mxDateTime.DateTime_Type, NULL, call);
+
+        /* if we found the above, we have this too. */
+        call = PyMapping_GetItemString(mod, "TimeFromMx");
+        microprotocols_add(mxDateTime.DateTimeDelta_Type, NULL, call);
+    }
+    else {
+        PyErr_Clear();
+    }
 #endif
 }
 
@@ -329,7 +249,32 @@ psyco_adapters_init(PyObject *mod)
    Fill the module's postgresql<->python encoding table */
 
 static encodingPair encodings[] = {
-    {"SQL_ASCII",    "ascii"},
+    {"ABC",          "cp1258"},
+    {"ALT",          "cp866"},
+    {"BIG5",         "big5"},
+    {"EUC_CN",       "euccn"},
+    {"EUC_JIS_2004", "euc_jis_2004"},
+    {"EUC_JP",       "euc_jp"},
+    {"EUC_KR",       "euc_kr"},
+    {"GB18030",      "gb18030"},
+    {"GBK",          "gbk"},
+    {"ISO_8859_1",   "iso8859_1"},
+    {"ISO_8859_2",   "iso8859_2"},
+    {"ISO_8859_3",   "iso8859_3"},
+    {"ISO_8859_5",   "iso8859_5"},
+    {"ISO_8859_6",   "iso8859_6"},
+    {"ISO_8859_7",   "iso8859_7"},
+    {"ISO_8859_8",   "iso8859_8"},
+    {"ISO_8859_9",   "iso8859_9"},
+    {"ISO_8859_10",  "iso8859_10"},
+    {"ISO_8859_13",  "iso8859_13"},
+    {"ISO_8859_14",  "iso8859_14"},
+    {"ISO_8859_15",  "iso8859_15"},
+    {"ISO_8859_16",  "iso8859_16"},
+    {"JOHAB",        "johab"},
+    {"KOI8",         "koi8_r"},
+    {"KOI8R",        "koi8_r"},
+    {"KOI8U",        "koi8_u"},
     {"LATIN1",       "iso8859_1"},
     {"LATIN2",       "iso8859_2"},
     {"LATIN3",       "iso8859_3"},
@@ -339,46 +284,30 @@ static encodingPair encodings[] = {
     {"LATIN7",       "iso8859_13"},
     {"LATIN8",       "iso8859_14"},
     {"LATIN9",       "iso8859_15"},
-    {"ISO88591",     "iso8859_1"},
-    {"ISO88592",     "iso8859_2"},
-    {"ISO88593",     "iso8859_3"},
-    {"ISO88595",     "iso8859_5"},
-    {"ISO88596",     "iso8859_6"},
-    {"ISO88597",     "iso8859_7"},
-    {"ISO885913",    "iso8859_13"},
-    {"ISO88598",     "iso8859_8"},
-    {"ISO88599",     "iso8859_9"},
-    {"ISO885914",    "iso8859_14"},
-    {"ISO885915",    "iso8859_15"},
-    {"UNICODE",      "utf_8"}, /* Not valid in 8.2, backward compatibility */
-    {"UTF8",         "utf_8"},
-    {"WIN950",       "cp950"},
-    {"Windows950",   "cp950"},
-    {"BIG5",         "big5"},
-    {"EUC_JP",       "euc_jp"},
-    {"EUC_KR",       "euc_kr"},
-    {"GB18030",      "gb18030"},
-    {"GBK",          "gbk"},
-    {"WIN936",       "gbk"},
-    {"Windows936",   "gbk"},
-    {"JOHAB",        "johab"},
-    {"KOI8",         "koi8_r"}, /* in PG: KOI8 == KOI8R == KOI8-R == KOI8-U
-                                   but in Python there is koi8_r AND koi8_u */
-    {"KOI8R",        "koi8_r"},
-    {"SJIS",         "cp932"},
+    {"LATIN10",      "iso8859_16"},
     {"Mskanji",      "cp932"},
     {"ShiftJIS",     "cp932"},
-    {"WIN932",       "cp932"},
-    {"Windows932",   "cp932"},
+    {"SHIFT_JIS_2004", "shift_jis_2004"},
+    {"SJIS",         "cp932"},
+    {"SQL_ASCII",    "ascii"},  /* XXX this is wrong: SQL_ASCII means "no
+                                 *  encoding" we should fix the unicode
+                                 *  typecaster to return a str or bytes in Py3
+                                 */
+    {"TCVN",         "cp1258"},
+    {"TCVN5712",     "cp1258"},
     {"UHC",          "cp949"},
-    {"WIN949",       "cp949"},
-    {"Windows949",   "cp949"},
+    {"UNICODE",      "utf_8"}, /* Not valid in 8.2, backward compatibility */
+    {"UTF8",         "utf_8"},
+    {"VSCII",        "cp1258"},
+    {"WIN",          "cp1251"},
     {"WIN866",       "cp866"},
-    {"ALT",          "cp866"},
     {"WIN874",       "cp874"},
+    {"WIN932",       "cp932"},
+    {"WIN936",       "gbk"},
+    {"WIN949",       "cp949"},
+    {"WIN950",       "cp950"},
     {"WIN1250",      "cp1250"},
     {"WIN1251",      "cp1251"},
-    {"WIN",          "cp1251"},
     {"WIN1252",      "cp1252"},
     {"WIN1253",      "cp1253"},
     {"WIN1254",      "cp1254"},
@@ -386,16 +315,13 @@ static encodingPair encodings[] = {
     {"WIN1256",      "cp1256"},
     {"WIN1257",      "cp1257"},
     {"WIN1258",      "cp1258"},
-    {"ABC",          "cp1258"},
-    {"TCVN",         "cp1258"},
-    {"TCVN5712",     "cp1258"},
-    {"VSCII",        "cp1258"},
+    {"Windows932",   "cp932"},
+    {"Windows936",   "gbk"},
+    {"Windows949",   "cp949"},
+    {"Windows950",   "cp950"},
 
 /* those are missing from Python:                */
-/*    {"EUC_CN", "?"},                           */
 /*    {"EUC_TW", "?"},                           */
-/*    {"LATIN10", "?"},                          */
-/*    {"ISO885916", "?"},                        */
 /*    {"MULE_INTERNAL", "?"},                    */
     {NULL, NULL}
 };
@@ -405,7 +331,7 @@ static void psyco_encodings_fill(PyObject *dict)
     encodingPair *enc;
 
     for (enc = encodings; enc->pgenc != NULL; enc++) {
-        PyObject *value = PyString_FromString(enc->pyenc);
+        PyObject *value = Text_FromUTF8(enc->pyenc);
         PyDict_SetItemString(dict, enc->pgenc, value);
         Py_DECREF(value);
     }
@@ -470,12 +396,18 @@ psyco_errors_init(void)
         dict = PyDict_New();
 
         if (exctable[i].docstr) {
-            str = PyString_FromString(exctable[i].docstr);
+            str = Text_FromUTF8(exctable[i].docstr);
             PyDict_SetItemString(dict, "__doc__", str);
         }
 
-        if (exctable[i].base == 0)
+        if (exctable[i].base == 0) {
+            #if PY_MAJOR_VERSION < 3
             base = PyExc_StandardError;
+            #else
+            /* StandardError is gone in 3.0 */
+            base = NULL;
+            #endif
+        }
         else
             base = *exctable[i].base;
 
@@ -535,31 +467,47 @@ psyco_errors_set(PyObject *type)
    Create a new error of the given type with extra attributes. */
 
 void
-psyco_set_error(PyObject *exc, PyObject *curs, const char *msg,
+psyco_set_error(PyObject *exc, cursorObject *curs, const char *msg,
                 const char *pgerror, const char *pgcode)
 {
     PyObject *t;
+    PyObject *pymsg;
+    PyObject *err = NULL;
+    connectionObject *conn = NULL;
 
-    PyObject *err = PyObject_CallFunction(exc, "s", msg);
+    if (curs) {
+        conn = ((cursorObject *)curs)->conn;
+    }
+
+    if ((pymsg = conn_text_from_chars(conn, msg))) {
+        err = PyObject_CallFunctionObjArgs(exc, pymsg, NULL);
+        Py_DECREF(pymsg);
+    }
+    else {
+        /* what's better than an error in an error handler in the morning?
+         * Anyway, some error was set, refcount is ok... get outta here. */
+        return;
+    }
 
     if (err) {
+        if (curs) {
+            PyObject_SetAttrString(err, "cursor", (PyObject *)curs);
+        }
+
         if (pgerror) {
-            t = PyString_FromString(pgerror);
+            t = conn_text_from_chars(conn, pgerror);
             PyObject_SetAttrString(err, "pgerror", t);
             Py_DECREF(t);
         }
 
         if (pgcode) {
-            t = PyString_FromString(pgcode);
+            t = conn_text_from_chars(conn, pgcode);
             PyObject_SetAttrString(err, "pgcode", t);
             Py_DECREF(t);
         }
 
-        if (curs)
-            PyObject_SetAttrString(err, "cursor", curs);
-
         PyErr_SetObject(exc, err);
-    Py_DECREF(err);
+        Py_DECREF(err);
     }
 }
 
@@ -631,10 +579,48 @@ psyco_GetDecimalType(void)
 }
 
 
+/* Create a namedtuple for cursor.description items
+ *
+ * Return None in case of expected errors (e.g. namedtuples not available)
+ * NULL in case of errors to propagate.
+ */
+static PyObject *
+psyco_make_description_type(void)
+{
+    PyObject *nt = NULL;
+    PyObject *coll = NULL;
+    PyObject *rv = NULL;
+
+    /* Try to import collections.namedtuple */
+    if (!(coll = PyImport_ImportModule("collections"))) {
+        Dprintf("psyco_make_description_type: collections import failed");
+        PyErr_Clear();
+        rv = Py_None;
+        goto exit;
+    }
+    if (!(nt = PyObject_GetAttrString(coll, "namedtuple"))) {
+        Dprintf("psyco_make_description_type: no collections.namedtuple");
+        PyErr_Clear();
+        rv = Py_None;
+        goto exit;
+    }
+
+    /* Build the namedtuple */
+    rv = PyObject_CallFunction(nt, "ss", "Column",
+        "name type_code display_size internal_size precision scale null_ok");
+
+exit:
+    Py_XDECREF(coll);
+    Py_XDECREF(nt);
+
+    return rv;
+}
+
+
 /** method table and module initialization **/
 
 static PyMethodDef psycopgMethods[] = {
-    {"connect",  (PyCFunction)psyco_connect,
+    {"_connect",  (PyCFunction)psyco_connect,
      METH_VARARGS|METH_KEYWORDS, psyco_connect_doc},
     {"adapt",  (PyCFunction)psyco_microprotocols_adapt,
      METH_VARARGS, psyco_microprotocols_adapt_doc},
@@ -643,17 +629,21 @@ static PyMethodDef psycopgMethods[] = {
      METH_VARARGS, psyco_register_type_doc},
     {"new_type", (PyCFunction)typecast_from_python,
      METH_VARARGS|METH_KEYWORDS, typecast_from_python_doc},
+    {"new_array_type", (PyCFunction)typecast_array_from_python,
+     METH_VARARGS|METH_KEYWORDS, typecast_array_from_python_doc},
 
     {"AsIs",  (PyCFunction)psyco_AsIs,
      METH_VARARGS, psyco_AsIs_doc},
     {"QuotedString",  (PyCFunction)psyco_QuotedString,
      METH_VARARGS, psyco_QuotedString_doc},
     {"Boolean",  (PyCFunction)psyco_Boolean,
-     METH_VARARGS, psyco_Float_doc},
-    {"Float",  (PyCFunction)psyco_Float,
-     METH_VARARGS, psyco_Decimal_doc},
-    {"Decimal",  (PyCFunction)psyco_Decimal,
      METH_VARARGS, psyco_Boolean_doc},
+    {"Int",  (PyCFunction)psyco_Int,
+     METH_VARARGS, psyco_Int_doc},
+    {"Float",  (PyCFunction)psyco_Float,
+     METH_VARARGS, psyco_Float_doc},
+    {"Decimal",  (PyCFunction)psyco_Decimal,
+     METH_VARARGS, psyco_Decimal_doc},
     {"Binary",  (PyCFunction)psyco_Binary,
      METH_VARARGS, psyco_Binary_doc},
     {"Date",  (PyCFunction)psyco_Date,
@@ -681,6 +671,7 @@ static PyMethodDef psycopgMethods[] = {
      METH_VARARGS, psyco_IntervalFromPy_doc},
 
 #ifdef HAVE_MXDATETIME
+    /* to be deleted if not found at import time */
     {"DateFromMx",  (PyCFunction)psyco_DateFromMx,
      METH_VARARGS, psyco_DateFromMx_doc},
     {"TimeFromMx",  (PyCFunction)psyco_TimeFromMx,
@@ -701,13 +692,29 @@ static PyMethodDef psycopgMethods[] = {
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
-PyMODINIT_FUNC
-init_psycopg(void)
-{
-    static void *PSYCOPG_API[PSYCOPG_API_pointers];
+#if PY_MAJOR_VERSION > 2
+static struct PyModuleDef psycopgmodule = {
+        PyModuleDef_HEAD_INIT,
+        "_psycopg",
+        NULL,
+        -1,
+        psycopgMethods,
+        NULL,
+        NULL,
+        NULL,
+        NULL
+};
+#endif
 
-    PyObject *module, *dict;
+PyMODINIT_FUNC
+INIT_MODULE(_psycopg)(void)
+{
+#if PY_VERSION_HEX < 0x03020000
+    static void *PSYCOPG_API[PSYCOPG_API_pointers];
     PyObject *c_api_object;
+#endif
+
+    PyObject *module = NULL, *dict;
 
 #ifdef PSYCOPG_DEBUG
     if (getenv("PSYCOPG_DEBUG"))
@@ -717,51 +724,57 @@ init_psycopg(void)
     Dprintf("initpsycopg: initializing psycopg %s", PSYCOPG_VERSION);
 
     /* initialize all the new types and then the module */
-    connectionType.ob_type = &PyType_Type;
-    cursorType.ob_type     = &PyType_Type;
-    typecastType.ob_type   = &PyType_Type;
-    qstringType.ob_type    = &PyType_Type;
-    binaryType.ob_type     = &PyType_Type;
-    isqlquoteType.ob_type  = &PyType_Type;
-    pbooleanType.ob_type   = &PyType_Type;
-    pfloatType.ob_type     = &PyType_Type;
-    pdecimalType.ob_type   = &PyType_Type;
-    asisType.ob_type       = &PyType_Type;
-    listType.ob_type       = &PyType_Type;
-    chunkType.ob_type      = &PyType_Type;
-    NotifyType.ob_type     = &PyType_Type;
-    XidType.ob_type        = &PyType_Type;
+    Py_TYPE(&connectionType) = &PyType_Type;
+    Py_TYPE(&cursorType)     = &PyType_Type;
+    Py_TYPE(&typecastType)   = &PyType_Type;
+    Py_TYPE(&qstringType)    = &PyType_Type;
+    Py_TYPE(&binaryType)     = &PyType_Type;
+    Py_TYPE(&isqlquoteType)  = &PyType_Type;
+    Py_TYPE(&pbooleanType)   = &PyType_Type;
+    Py_TYPE(&pintType)       = &PyType_Type;
+    Py_TYPE(&pfloatType)     = &PyType_Type;
+    Py_TYPE(&pdecimalType)   = &PyType_Type;
+    Py_TYPE(&asisType)       = &PyType_Type;
+    Py_TYPE(&listType)       = &PyType_Type;
+    Py_TYPE(&chunkType)      = &PyType_Type;
+    Py_TYPE(&NotifyType)     = &PyType_Type;
+    Py_TYPE(&XidType)        = &PyType_Type;
 
-    if (PyType_Ready(&connectionType) == -1) return;
-    if (PyType_Ready(&cursorType) == -1) return;
-    if (PyType_Ready(&typecastType) == -1) return;
-    if (PyType_Ready(&qstringType) == -1) return;
-    if (PyType_Ready(&binaryType) == -1) return;
-    if (PyType_Ready(&isqlquoteType) == -1) return;
-    if (PyType_Ready(&pbooleanType) == -1) return;
-    if (PyType_Ready(&pfloatType) == -1) return;
-    if (PyType_Ready(&pdecimalType) == -1) return;
-    if (PyType_Ready(&asisType) == -1) return;
-    if (PyType_Ready(&listType) == -1) return;
-    if (PyType_Ready(&chunkType) == -1) return;
-    if (PyType_Ready(&NotifyType) == -1) return;
-    if (PyType_Ready(&XidType) == -1) return;
+    if (PyType_Ready(&connectionType) == -1) goto exit;
+    if (PyType_Ready(&cursorType) == -1) goto exit;
+    if (PyType_Ready(&typecastType) == -1) goto exit;
+    if (PyType_Ready(&qstringType) == -1) goto exit;
+    if (PyType_Ready(&binaryType) == -1) goto exit;
+    if (PyType_Ready(&isqlquoteType) == -1) goto exit;
+    if (PyType_Ready(&pbooleanType) == -1) goto exit;
+    if (PyType_Ready(&pintType) == -1) goto exit;
+    if (PyType_Ready(&pfloatType) == -1) goto exit;
+    if (PyType_Ready(&pdecimalType) == -1) goto exit;
+    if (PyType_Ready(&asisType) == -1) goto exit;
+    if (PyType_Ready(&listType) == -1) goto exit;
+    if (PyType_Ready(&chunkType) == -1) goto exit;
+    if (PyType_Ready(&NotifyType) == -1) goto exit;
+    if (PyType_Ready(&XidType) == -1) goto exit;
 
 #ifdef PSYCOPG_EXTENSIONS
-    lobjectType.ob_type    = &PyType_Type;
-    if (PyType_Ready(&lobjectType) == -1) return;
+    Py_TYPE(&lobjectType)    = &PyType_Type;
+    if (PyType_Ready(&lobjectType) == -1) goto exit;
 #endif
 
     /* import mx.DateTime module, if necessary */
 #ifdef HAVE_MXDATETIME
-    mxdatetimeType.ob_type = &PyType_Type;
-    if (PyType_Ready(&mxdatetimeType) == -1) return;
-    if (mxDateTime_ImportModuleAndAPI() != 0) {
-        Dprintf("initpsycopg: why marc hide mx.DateTime again?!");
-        PyErr_SetString(PyExc_ImportError, "can't import mx.DateTime module");
-        return;
+    Py_TYPE(&mxdatetimeType) = &PyType_Type;
+    if (PyType_Ready(&mxdatetimeType) == -1) goto exit;
+    if (0 != mxDateTime_ImportModuleAndAPI()) {
+        PyErr_Clear();
+
+        /* only fail if the mx typacaster should have been the default */
+#ifdef PSYCOPG_DEFAULT_MXDATETIME
+        PyErr_SetString(PyExc_ImportError,
+            "can't import mx.DateTime module (requested as default adapter)");
+        goto exit;
+#endif
     }
-    if (psyco_adapter_mxdatetime_init()) { return; }
 #endif
 
     /* import python builtin datetime module, if available */
@@ -769,22 +782,22 @@ init_psycopg(void)
     if (pyDateTimeModuleP == NULL) {
         Dprintf("initpsycopg: can't import datetime module");
         PyErr_SetString(PyExc_ImportError, "can't import datetime module");
-        return;
+        goto exit;
     }
 
     /* Initialize the PyDateTimeAPI everywhere is used */
     PyDateTime_IMPORT;
-    if (psyco_adapter_datetime_init()) { return; }
+    if (psyco_adapter_datetime_init()) { goto exit; }
 
-    pydatetimeType.ob_type = &PyType_Type;
-    if (PyType_Ready(&pydatetimeType) == -1) return;
+    Py_TYPE(&pydatetimeType) = &PyType_Type;
+    if (PyType_Ready(&pydatetimeType) == -1) goto exit;
 
     /* import psycopg2.tz anyway (TODO: replace with C-level module?) */
     pyPsycopgTzModule = PyImport_ImportModule("psycopg2.tz");
     if (pyPsycopgTzModule == NULL) {
         Dprintf("initpsycopg: can't import psycopg2.tz module");
         PyErr_SetString(PyExc_ImportError, "can't import psycopg2.tz module");
-        return;
+        goto exit;
     }
     pyPsycopgTzLOCAL =
         PyObject_GetAttrString(pyPsycopgTzModule, "LOCAL");
@@ -792,27 +805,38 @@ init_psycopg(void)
         PyObject_GetAttrString(pyPsycopgTzModule, "FixedOffsetTimezone");
 
     /* initialize the module and grab module's dictionary */
+#if PY_MAJOR_VERSION < 3
     module = Py_InitModule("_psycopg", psycopgMethods);
+#else
+    module = PyModule_Create(&psycopgmodule);
+#endif
+    if (!module) { goto exit; }
+
     dict = PyModule_GetDict(module);
 
     /* initialize all the module's exported functions */
     /* PyBoxer_API[PyBoxer_Fake_NUM] = (void *)PyBoxer_Fake; */
 
     /* Create a CObject containing the API pointer array's address */
+    /* If anybody asks for a PyCapsule we'll deal with it. */
+#if PY_VERSION_HEX < 0x03020000
     c_api_object = PyCObject_FromVoidPtr((void *)PSYCOPG_API, NULL);
     if (c_api_object != NULL)
         PyModule_AddObject(module, "_C_API", c_api_object);
+#endif
 
     /* other mixed initializations of module-level variables */
     psycoEncodings = PyDict_New();
     psyco_encodings_fill(psycoEncodings);
+    psyco_null = Bytes_FromString("NULL");
+    psyco_DescriptionType = psyco_make_description_type();
 
     /* set some module's parameters */
     PyModule_AddStringConstant(module, "__version__", PSYCOPG_VERSION);
     PyModule_AddStringConstant(module, "__doc__", "psycopg PostgreSQL driver");
-    PyModule_AddObject(module, "apilevel", PyString_FromString(APILEVEL));
+    PyModule_AddObject(module, "apilevel", Text_FromUTF8(APILEVEL));
     PyModule_AddObject(module, "threadsafety", PyInt_FromLong(THREADSAFETY));
-    PyModule_AddObject(module, "paramstyle", PyString_FromString(PARAMSTYLE));
+    PyModule_AddObject(module, "paramstyle", Text_FromUTF8(PARAMSTYLE));
 
     /* put new types in module dictionary */
     PyModule_AddObject(module, "connection", (PyObject*)&connectionType);
@@ -827,6 +851,16 @@ init_psycopg(void)
     /* encodings dictionary in module dictionary */
     PyModule_AddObject(module, "encodings", psycoEncodings);
 
+#ifdef HAVE_MXDATETIME
+    /* If we can't find mx.DateTime objects at runtime,
+     * remove them from the module (and, as consequence, from the adapters). */
+    if (0 != psyco_adapter_mxdatetime_init()) {
+        PyDict_DelItemString(dict, "DateFromMx");
+        PyDict_DelItemString(dict, "TimeFromMx");
+        PyDict_DelItemString(dict, "TimestampFromMx");
+        PyDict_DelItemString(dict, "IntervalFromMx");
+    }
+#endif
     /* initialize default set of typecasters */
     typecast_init(dict);
 
@@ -843,6 +877,7 @@ init_psycopg(void)
     binaryType.tp_alloc = PyType_GenericAlloc;
     isqlquoteType.tp_alloc = PyType_GenericAlloc;
     pbooleanType.tp_alloc = PyType_GenericAlloc;
+    pintType.tp_alloc = PyType_GenericAlloc;
     pfloatType.tp_alloc = PyType_GenericAlloc;
     pdecimalType.tp_alloc = PyType_GenericAlloc;
     connectionType.tp_alloc = PyType_GenericAlloc;
@@ -858,10 +893,16 @@ init_psycopg(void)
     lobjectType.tp_alloc = PyType_GenericAlloc;
 #endif
 
-
 #ifdef HAVE_MXDATETIME
     mxdatetimeType.tp_alloc = PyType_GenericAlloc;
 #endif
 
     Dprintf("initpsycopg: module initialization complete");
+
+exit:
+#if PY_MAJOR_VERSION > 2
+    return module;
+#else
+    return;
+#endif
 }
